import os
import argparse
import pandas as pd
import numpy as np
from concurrent.futures import ProcessPoolExecutor
import random
from scipy.optimize import curve_fit

#**************************************
#****DESIGN BY Wang Taiquan*******
#**IF USE, PLEASE CITE THE ARTICLE**
#**************************************

def heaps_law(N, k, lam):
    return k * (N ** lam)


def calculate_lambda(gpa_df, genomes, sample_sizes, max_samples, repeats):
    lambda_values = []
    for _ in range(repeats):  
        median_gene_counts = []
        for sample_size in sample_sizes:
            gene_counts = []
            for _ in range(max_samples):
                sampled_genomes = random.sample(genomes, sample_size)
                non_redundant_gene_count = gpa_df[sampled_genomes].notnull().sum(axis=1).astype(bool).sum()
                gene_counts.append(non_redundant_gene_count)
            median_gene_counts.append(np.median(gene_counts))

        N = np.array(sample_sizes)
        P = np.array(median_gene_counts)
        try:
            params, _ = curve_fit(heaps_law, N, P, maxfev=10000)
            _, lam = params
            lambda_values.append(lam)
        except RuntimeError as e:
            print(f"Curve fitting failed for {sample_sizes}: {e}")
            continue

    if lambda_values:
        return np.mean(lambda_values), np.std(lambda_values), len(lambda_values)
    else:
        return np.nan, np.nan, 0



# Here is the function for processing individual species and returning the results.
def calculate_lambda_for_species(args):
    species_name, genome_count, file_path = args
    gpa_df = pd.read_csv(file_path, low_memory=False)
    genomes = list(gpa_df.columns[3:])  # Genome starting from column 4

    # Select sample size list based on number of genomes
    if genome_count > 40:
        sample_sizes = np.linspace(2, genome_count, 30).astype(int).tolist()
    elif genome_count > 1:
        sample_sizes = list(range(1, genome_count))
    else:
        return f"Insufficient genomes for {species_name}, skipping...\n"

    max_samples = 50
    repeats = 20
    mean_lambda, std_lambda, samples = calculate_lambda(gpa_df, genomes, sample_sizes, max_samples, repeats)
    return f"{species_name}\t{mean_lambda}\t{std_lambda}\t{samples}\n"


def main(cpus, species_info):
    # output file path
    result_file_path = "/path/to/name.txt"


    tasks = []
    for species_name, genome_count in species_info.items():
        file_path = f"/path/to/{species_name}/gene_presence_absence.csv"   #means input data csv
        if os.path.exists(file_path):
            tasks.append((species_name, genome_count, file_path))
        else:
            print(f"File not found for {species_name}, skipping...")

    # the number of cpu
    with ProcessPoolExecutor(max_workers=cpus) as executor:
        results = list(executor.map(calculate_lambda_for_species, tasks))

    with open(result_file_path, 'w') as result_file:
        for result in results:
            result_file.write(result)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Parallel lambda calculation.')
    parser.add_argument('-c', '--cpus', type=int, default=os.cpu_count(), help='Number of CPUs to use for parallel computation.')
    args = parser.parse_args()

# Define species information, taking the data from this study as an example
    species_info = {
        "Bacillus_amyloliquefaciens": 176,
        "Bacillus_atrophaeus": 93,
        "Bacillus_halotolerans": 56,
        "Bacillus_inaquosorum": 106,
        "Bacillus_licheniformis": 274,
        "Bacillus_mojavensis": 11,
        "Bacillus_paralicheniformis": 136,
        "Bacillus_siamensis": 11,
        "Bacillus_sonorensis": 26,
        "Bacillus_spizizenii": 140,
        "Bacillus_stercoris": 8,
        "Bacillus_subtilis": 571,
        "Bacillus_tequilensis": 5,
        "Bacillus_vallismortis": 13,
        "Bacillus_velezensis": 446
    }

    main(args.cpus, species_info)
